from .base import BaseChatbot

class Chatbot_llama(BaseChatbot):
    def __init__(self, model_path='models/llama-2-7b.Q4_K_M.gguf', device='cuda', sys_prompt='',
                 break_words=None, name='[JOHN]'):
        from llama_cpp import Llama
        if break_words is None:
            break_words = ['[USER]', '[END]']
        self.model = Llama(model_path=model_path, n_ctx=4096,
                           n_gpu_layers=-1 if device == 'cuda' else 0,
                           verbose=False)
        self.sys_prompt = sys_prompt
        self.break_words = break_words
        self.name = name
        self.messages = []
        self.messages.append(self.sys_prompt)

    def run(self, input_text):
        self.messages.append(input_text + '\n' + self.name)
        out = self.model.create_completion(''.join(self.messages),
                                           max_tokens=1000, stream=True)
        response_text = ''
        for o in out:
            text = o['choices'][0]['text']
            response_text += text
            yield text
            if any([response_text.strip().endswith(break_word) for break_word in self.break_words]):
                break

    def post_process(self, response):
        self.messages.append(response)
        return response


if __name__ == "__main__":
    preprompt = call_pre_prompt
    john = Chatbot_llama(model_path='../models/llama-2-7b.Q4_K_M.gguf', sys_prompt=preprompt)
    print("type: exit, quit or stop to end the chat")
    print("Chat started:")
    while True:
        user_input = input(" ")
        if user_input.lower() in ["exit", "quit", "stop"]:
            break

        response = john.generate_response(user_input)

        print(response)



import sounddevice as sd
import re
from time import monotonic
import queue
import threading
from typing import Callable
import numpy as np
import inspect
import asyncio


def remove_words_in_brackets_and_spaces(text):
    '''
    :param text: input text
    :return: input text with the extra spaces and words in brackets removed. (e.g. [USER])
    '''
    pattern = r'\s*\[.*?\]\s*'
    cleaned_text = re.sub(pattern, ' ', text)
    cleaned_text = cleaned_text.strip()
    return cleaned_text


class BaseMouth:
    def __init__(self, sample_rate: int, player=sd):
        self.sample_rate = sample_rate
        self.sentence_stop_pattern = r'[.?](?=\s+\S)'
        self.interrupted = ''
        self.player = player

    def run_tts(self, text: str) -> np.ndarray:
        '''
        :param text: The text to synthesize speech for
        :return: audio numpy array for sounddevice
        '''
        raise NotImplementedError('This method should be implemented by the subclass')

    def say_text(self, text: str):
        '''
        :param text: The text to synthesize speech for
        calls run_tts and plays the audio using sounddevice.
        '''
        output = self.run_tts(text)
        self.player.play(output, samplerate=self.sample_rate)
        self.player.wait()

    def say(self, audio_queue: queue.Queue, listen_interruption_func: Callable):
        '''
        :param audio_queue: The queue where the audio is stored for it to be played
        :param listen_interruption_func: callable function from the ear class.
        Plays the audios in the queue using sounddevice. Stops if interruption occurred.
        '''
        self.interrupted = ''
        while True:
            output, text = audio_queue.get()
            if output is None:
                break
            # get the duration of audio
            duration = len(output) / self.sample_rate
            self.player.play(output, samplerate=self.sample_rate)
            interruption = listen_interruption_func(duration)
            if interruption:
                self.player.stop()
                self.interrupted = (interruption, text)
                break
            else:
                self.player.wait()

    def say_multiple(self, text: str, listen_interruption_func: Callable):
        '''
        :param text: Intput text to synthesize
        :param listen_interruption_func: callable function from the ear class
        Splits the text into sentences separated by ['.', '?', '!']. Then plays the sentences one by one
        using run_tts and say
        '''
        pattern = r'[.?!]'
        sentences = re.split(pattern, text)
        sentences = [sentence.strip() for sentence in sentences if sentence.strip()]
        print(sentences)
        audio_queue = queue.Queue()
        say_thread = threading.Thread(target=self.say, args=(audio_queue, listen_interruption_func))
        say_thread.start()
        for sentence in sentences:
            output = self.run_tts(sentence)
            audio_queue.put((output, sentence))
            if self.interrupted:
                break
        audio_queue.put((None, ''))
        say_thread.join()

    def _handle_interruption(self, responses_list, interrupt_queue):
        interrupt_transcription, interrupt_text = self.interrupted
        idx = responses_list.index(interrupt_text)
        assert idx != -1, "Interrupted text not found in responses list. This should not happen. Raise an issue."
        responses_list = responses_list[:idx] + ['...']
        interrupt_queue.put(interrupt_transcription)
        return responses_list

    def say_multiple_stream(self, text_queue: queue.Queue,
                            listen_interruption_func: Callable,
                            interrupt_queue: queue.Queue,
                            audio_queue: queue.Queue = None):
        '''
        :param text_queue: The queue where the llm adds the predicted tokens
        :param listen_interruption_func: callable function from the ear class
        :param interrupt_queue: The queue where True is put when interruption occurred.
        :param audio_queue: The queue where the audio to be played is placed
        Receives text from the text_queue. As soon as a sentence is made run_tts is called to
        synthesize its speech.
        '''
        response = ''
        all_response = []
        interrupt_text_list = []
        if audio_queue is None:
            audio_queue = queue.Queue()
        say_thread = threading.Thread(target=self.say, args=(audio_queue, listen_interruption_func))
        say_thread.start()
        while True:
            text = text_queue.get()
            if text is None:
                sentence = response
            else:
                response += text
                if bool(re.search(self.sentence_stop_pattern, response)):
                    sentences = re.split(self.sentence_stop_pattern, response, maxsplit=1)
                    sentence = sentences[0]
                    response = sentences[1]
                else:
                    continue
            if sentence.strip() == '':
                break
            clean_sentence = remove_words_in_brackets_and_spaces(sentence).strip()
            output = self.run_tts(clean_sentence)
            audio_queue.put((output, clean_sentence))
            all_response.append(sentence)
            interrupt_text_list.append(clean_sentence)
            if self.interrupted:
                all_response = self._handle_interruption(interrupt_text_list, interrupt_queue)
                self.interrupted = ''
                break
            if text is None:
                break
        audio_queue.put((None, ''))
        say_thread.join()
        if self.interrupted:
            all_response = self._handle_interruption(interrupt_text_list, interrupt_queue)
        text_queue.queue.clear()
        text_queue.put('. '.join(all_response))



import torch
from .utils import record_user, record_interruption, record_user_stream
from .vad import VoiceActivityDetection
import re
from time import monotonic
import numpy as np
from threading import Thread
from queue import Queue

class BaseEar:
    def __init__(self, silence_seconds=3,
                 not_interrupt_words=None,
                 listener=None):
        if not_interrupt_words is None:
            not_interrupt_words = ['you', 'yes', 'yeah', 'hmm']  # you because whisper says "you" in silence
        self.silence_seconds = silence_seconds
        self.not_interrupt_words = not_interrupt_words
        self.vad = VoiceActivityDetection()
        self.listener = listener

    @torch.no_grad()
    def transcribe(self, input: np.ndarray) -> str:
        '''
        :param input: fp32 numpy array of the audio
        :return: transcription
        '''
        raise NotImplementedError("This method should be implemented by the subclass")

    def transcribe_stream(self, audio_queue: Queue, transcription_queue: Queue):
        '''
        :param audio_queue: Queue containing audio chunks from pyaudio stream
        :param transcription_queue: Queue to put transcriptions
        '''
        raise NotImplementedError("This method should be implemented by the subclass")

    def listen(self) -> str:
        '''
        :return: transcription
        records audio using record_user and returns its transcription
        '''
        audio = record_user(self.silence_seconds, self.vad, self.listener)
        text = self.transcribe(audio)
        return text

    def listen_stream(self) -> str:
        '''
        :return: transcription
        records audio using record_user and returns its transcription
        '''
        audio_queue = Queue()
        transcription_queue = Queue()

        audio_thread = Thread(target=record_user_stream, args=(self.silence_seconds, self.vad, audio_queue))
        transcription_thread = Thread(target=self.transcribe_stream, args=(audio_queue, transcription_queue))

        audio_thread.start()
        transcription_thread.start()

        audio_thread.join()
        transcription_thread.join()

        text = ''
        while True:
            _ = transcription_queue.get()
            if _ is None:
                break
            text += _ + ' '
        return text

    def listen_timing(self):
        audio = record_user(self.silence_seconds, self.vad)
        start = monotonic()
        text = self.transcribe(audio)
        end = monotonic()
        return text, end - start

    def interrupt_listen(self, record_seconds=100) -> str:
        '''
        :param record_seconds: Max seconds to record for
        :return: boolean indicating the if an interruption occured
        Records audio with interruption. Transcribes audio if
        voice activity detected and returns True if transcription indicates
        interruption.
        '''
        while record_seconds > 0:
            interruption_audio = record_interruption(self.vad, record_seconds, streamer=self.listener)
            # duration of interruption audio
            if interruption_audio is None:
                return ''
            else:
                duration = len(interruption_audio) / 16_000
                text = self.transcribe(interruption_audio)
                # remove any punctuation using re
                text = re.sub(r'[^\w\s]', '', text)
                text = text.lower()
                text = text.strip()
                if text in self.not_interrupt_words:
                    record_seconds -= duration
                else:
                    return text
